"""Ingest dbt metadata into Luma."""

from enum import Enum
from pathlib import Path
import time

from rich import print
from rich.console import Console
from rich.panel import Panel
import typer

from luma.models.dbt import DBTManifest
from luma.utils import (
    check_ingestion_results,
    check_ingestion_status,
    json_to_dict,
    perform_ingestion_request,
)
from luma.utils.options import (
    DryRun,
    Follow,
    LumaURL,
    MetadataDir,
)


app = typer.Typer(no_args_is_help=True, pretty_exceptions_show_locals=False)
console = Console()


class IngestionStatus(Enum):
    successful = 0
    failed = 1
    pending = 2


@app.command()
def ingest(  # noqa: C901
    metadata_dir: Path = MetadataDir,
    luma_url: str = LumaURL,
    dry_run: bool = DryRun,
    follow: bool = Follow,
) -> None:
    """Ingest a bundle of JSON files into Luma.

    The files (manifest.json, catalog.json, sources.json, run_results.json) should be
    located in the specified directory.

    manifest.json and catalog.json are required.

    Uses the current working directory if 'metadata_dir' is not specified.
    """
    # Define JSON paths.
    manifest_json_path = metadata_dir / "manifest.json"
    catalog_json_path = metadata_dir / "catalog.json"
    sources_json_path = metadata_dir / "sources.json"
    run_results_json_path = metadata_dir / "run_results.json"

    # Ensure required files exist.
    if not manifest_json_path.is_file():
        print(Panel(f"[red]{manifest_json_path.absolute()} is not a file[/red]"))
        raise typer.Exit(1)
    if not catalog_json_path.is_file():
        print(Panel(f"[red]{catalog_json_path.absolute()} is not a file[/red]"))
        raise typer.Exit(1)

    # Convert each JSON to dict.
    manifest = json_to_dict(json_path=manifest_json_path)
    catalog = json_to_dict(json_path=catalog_json_path)
    sources = json_to_dict(json_path=sources_json_path)
    run_results = json_to_dict(json_path=run_results_json_path)

    # Validate the manifest.
    DBTManifest.validate(manifest)

    # Bundle the artifacts.
    artifacts_bundle = {
        "manifest_json": manifest,
        "catalog_json": catalog,
        "sources_json": sources,
        "run_results_json": run_results,
    }

    # If in dry run mode, print the bundle and exit.
    if dry_run:
        print(artifacts_bundle)
        raise typer.Exit(0)

    # Send ingestion request.
    endpoint = f"{luma_url}/api/v1/dbt/"
    response, ingestion_uuid = perform_ingestion_request(
        url=endpoint,
        method="POST",
        payload=artifacts_bundle,
        verify=False,
    )
    if not response.ok:
        raise typer.Exit(1)

    # Wait until ingestion is complete.
    if follow and ingestion_uuid:
        ingestion_status = None

        with console.status("Waiting...", spinner="dots"):
            for _ in range(30):
                ingestion_status = check_ingestion_status(
                    luma_url, ingestion_uuid, verify=False
                )
                if ingestion_status == IngestionStatus.successful.value:
                    response = check_ingestion_results(luma_url, ingestion_uuid)
                    print()
                    print(f"Ingestion results for ID {ingestion_uuid}:")
                    print()
                    print(response)
                    return

                if ingestion_status == IngestionStatus.failed.value:
                    print()
                    print(f"Ingestion failed for ID {ingestion_uuid}")
                    return

                if ingestion_status == IngestionStatus.pending.value:
                    time.sleep(1)

        if ingestion_status != IngestionStatus.successful.value:
            print(
                f"Ingestion did not complete successfully within the wait period. Status: {ingestion_status}"
            )


@app.command()
def send_test_results(
    metadata_dir: Path = MetadataDir,
    luma_url: str = LumaURL,
    dry_run: bool = DryRun,
    follow: bool = Follow,
) -> None:
    """Send 'run_results.json' file located in the specified directory to Luma.

    The command will fail if the 'run_results.json' file is not present in metadata_dir.
    The current working directory is used if 'metadata_dir' is not specified.
    """
    run_results_path = Path(metadata_dir) / "run_results.json"
    run_results_dict = json_to_dict(json_path=run_results_path)

    # If in dry run mode, print the test results and exit.
    if dry_run:
        print(run_results_dict)
        raise typer.Exit(0)

    # Send ingestion request.
    endpoint = f"{luma_url}/api/v1/dbt/run_results/"
    response, ingestion_uuid = perform_ingestion_request(
        url=endpoint,
        method="POST",
        payload=run_results_dict,
        verify=False,
    )
    if not response.ok:
        raise typer.Exit(1)

    if follow and ingestion_uuid:
        ingestion_status = None

        with console.status("Waiting...", spinner="dots"):
            for _ in range(30):
                ingestion_status = check_ingestion_status(luma_url, ingestion_uuid)
                if ingestion_status == IngestionStatus.successful.value:
                    response = check_ingestion_results(luma_url, ingestion_uuid)
                    print()
                    print(f"Ingestion results for ID {ingestion_uuid}:")
                    print()
                    print(response)
                    return

                if ingestion_status == IngestionStatus.failed.value:
                    print()
                    print(f"Ingestion failed for ID {ingestion_uuid}")
                    return

                if ingestion_status == IngestionStatus.pending.value:
                    time.sleep(1)

        if ingestion_status != IngestionStatus.successful.value:
            print(
                f"Ingestion did not complete successfully within the wait period. Status: {ingestion_status}"
            )


# Run the application.
if __name__ == "__main__":
    app()



import pytest

from luma.utils import (
    create_conn,
    generate_pg_dump_content,
    get_db_metadata,
    get_pg_dump_tables_info,
    get_pg_dump_views_info,
    get_tables_row_counts,
    get_tables_size_info,
    run_command,
)
from unittest import skip


@pytest.mark.skip()
def test_run_command():
    """
    Test the `run_command` function to ensure it correctly executes shell commands
    and returns the output.
    """
    command = 'echo "Hello, World!"'
    assert run_command(command, True) == "Hello, World!"


@pytest.mark.skip()
def test_create_conn(setup_db):
    """
    Test the `create_conn` function to ensure it creates a database connection 
    using provided credentials and connection information.
    """
    assert isinstance(
        setup_db.info.port, int
    )  # Add this line to ensure port is an integer

    conn = create_conn(
        username=setup_db.info.user,
        password=setup_db.info.password,
        host=setup_db.info.host,
        port=setup_db.info.port,
        database=setup_db.info.dbname,
    )
    assert conn
    assert conn.closed == 0  # 0 indicates connection is open


@pytest.mark.skip()
def test_generate_pg_dump_content(setup_db):
    """
    Test the `generate_pg_dump_content` function to ensure it correctly generates 
    PostgreSQL database dump content.
    """
    result = generate_pg_dump_content(
        username=setup_db.info.user,
        password=setup_db.info.password,
        host=setup_db.info.host,
        port=setup_db.info.port,
        database=setup_db.info.dbname,
    )
    assert "users" in result
    assert "products" in result
    assert "product_id" in result
    assert "user_id" in result
    assert "CURRENT_TIMESTAMP" in result


@pytest.mark.skip()
def test_get_pg_dump_tables_info(setup_db):
    """
    Test the `get_pg_dump_tables_info` function to validate extraction of table 
    information from PostgreSQL dump content.
    """
    dump_content = generate_pg_dump_content(
        username=setup_db.info.user,
        password=setup_db.info.password,
        host=setup_db.info.host,
        port=setup_db.info.port,
        database=setup_db.info.dbname,
    )
    result = get_pg_dump_tables_info(dump_content)
    assert isinstance(result, list)
    assert len(result) > 0


@pytest.mark.skip()
def test_get_pg_dump_views_info(setup_db):
    """
    Test the `get_pg_dump_views_info` function to validate extraction of view 
    information from PostgreSQL dump content.
    """
    dump_content = generate_pg_dump_content(
        username=setup_db.info.user,
        password=setup_db.info.password,
        host=setup_db.info.host,
        port=setup_db.info.port,
        database=setup_db.info.dbname,
    )
    result = get_pg_dump_views_info(dump_content)
    assert isinstance(result, list)
    assert len(result) > 0


@pytest.mark.skip()
def test_get_tables_size_info(setup_db):
    """
    Test the `get_tables_size_info` function to ensure it retrieves information 
    about the size of tables in a database.
    """
    result = get_tables_size_info(
        username=setup_db.info.user,
        password=setup_db.info.password,
        host=setup_db.info.host,
        port=setup_db.info.port,
        database=setup_db.info.dbname,
    )
    assert isinstance(result, dict)
    assert len(result) > 0


@pytest.mark.skip()
def test_get_tables_row_counts(setup_db):
    """
    Test the `get_tables_row_counts` function to ensure it retrieves row counts 
    for each table in a database.
    """
    result = get_tables_row_counts(
        username=setup_db.info.user,
        password=setup_db.info.password,
        host=setup_db.info.host,
        port=setup_db.info.port,
        database=setup_db.info.dbname,
    )
    assert isinstance(result, dict)
    assert len(result) > 0


@pytest.mark.skip()
def test_get_db_metadata(setup_db):
    """
    Test the `get_db_metadata` function to ensure it retrieves metadata for a 
    database.
    """
    result = get_db_metadata(
        username=setup_db.info.user,
        password=setup_db.info.password,
        host=setup_db.info.host,
        port=setup_db.info.port,
        database=setup_db.info.dbname,
    )
    assert isinstance(result, dict)
    assert len(result) > 0



import dlt


try:
    import pyrfc
    from pyrfc._exception import ABAPApplicationError, ABAPRuntimeError
except ModuleNotFoundError as e:
    msg = "The 'pyrfc' package is required to use the SAPRFC source. "
    raise ImportError(msg) from e
from itertools import batched
import string
import textwrap

from loguru import logger


@dlt.source(name="sap")
def sap(
    ashost: str = dlt.secrets.value,
    sysnr: str = dlt.secrets.value,
    username: str = dlt.secrets.value,
    passwd: str = dlt.secrets.value,
):
    """Query SAP with SQL using the RFC protocol.

    Use the RFC_READ_TABLE to read SAP table metadata.
    """
    delimiter = "\t"

    # Test the connection.
    con = pyrfc.Connection(
        ashost=ashost,
        sysnr=sysnr,
        user=username,
        passwd=passwd,
    )
    logger.info("Checking the connection...")
    try:
        con.ping()
        logger.info("Connection successful.")
    except Exception as e:
        logger.exception("Connection to SAP failed.")
        raise
    finally:
        con.close()

    # By convention, custom tables use a "Z" or "Y" prefix in their name.
    custom_table_prefixes = ["Z", "Y"]
    standard_table_prefixes = [
        char for char in string.ascii_uppercase if char not in custom_table_prefixes
    ]

    def get_response(params):
        """Call the RFC_READ_TABLE function with the given parameters."""
        con = pyrfc.Connection(
            ashost=ashost,
            sysnr=sysnr,
            user=username,
            passwd=passwd,
        )
        try:
            response = con.call("RFC_READ_TABLE", **params)
        except (ABAPApplicationError, ABAPRuntimeError) as e:
            if e.key == "DATA_BUFFER_EXCEEDED":
                msg = "Character limit per row exceeded. Please select fewer columns."
            elif e.key == "TSV_TNEW_PAGE_ALLOC_FAILED":
                msg = "Memory allocation failed; try a smaller batch size."
            else:
                msg = f"Error while calling RFC_READ_TABLE with params:\n{params}."
            logger.exception(msg)
            raise
        finally:
            con.close()

        # Process the response into records.
        record_key = "WA"
        data_raw = response["DATA"]
        data = [
            [value.strip() for value in row[record_key].split(delimiter)]
            for row in data_raw
        ]
        logger.info(f"Retrieved {len(data)} rows.")
        columns = params["FIELDS"]
        return [dict(zip(columns, row)) for row in data]

    def get_table_schema(table_prefixes: list[str]):
        conditions = [f"TABNAME LIKE '{prefix}%'" for prefix in table_prefixes]
        # Each line in OPTIONS must be ≤72 characters.
        grouped = textwrap.wrap(" OR ".join(conditions), width=72)
        options = [{"TEXT": line} for line in grouped]
        # Filter out internal SAP objects.
        options += [{"TEXT": " AND TABCLASS = 'TRANSP'"}]
        params = {
            "QUERY_TABLE": "DD02L",
            "FIELDS": ["TABNAME"],
            "OPTIONS": options,
            "DELIMITER": delimiter,
        }
        yield get_response(params)

    @dlt.resource(
        name="standard_tables", write_disposition="merge", primary_key="TABNAME"
    )
    def standard_tables():
        """Get a list of standard SAP tables."""
        yield get_table_schema(standard_table_prefixes)

    @dlt.resource(write_disposition="merge", primary_key="TABNAME")
    def custom_tables():
        """Get a list of custom SAP tables."""
        yield get_table_schema(custom_table_prefixes)

    def get_table_details(table_names):
        """Get metadata about SAP tables."""
        table_batches = batched(table_names, 1000)
        for batch_number, batch in enumerate(table_batches, start=1):
            conditions = [f"TABNAME = '{table}'" for table in batch]
            # Each line in OPTIONS must be ≤72 characters.
            grouped = textwrap.wrap(" OR ".join(conditions), width=72)
            options = [{"TEXT": line} for line in grouped]
            params = {
                "QUERY_TABLE": "DD03L",
                "FIELDS": ["TABNAME", "FIELDNAME", "DATATYPE", "LENG"],
                "OPTIONS": options,
                "DELIMITER": delimiter,
            }
            logger.info(f"Extracting table batch number {batch_number}...")
            yield get_response(params)

    @dlt.transformer(
        data_from=standard_tables,
        write_disposition="merge",
        primary_key=("TABNAME", "FIELDNAME"),
    )
    def standard_tables_details(standard_tables):
        standard_table_names = [row["TABNAME"] for row in standard_tables]
        yield get_table_details(standard_table_names)

    @dlt.transformer(
        data_from=custom_tables,
        write_disposition="merge",
        primary_key=("TABNAME", "FIELDNAME"),
    )
    def custom_tables_details(custom_tables):
        custom_table_names = [row["TABNAME"] for row in custom_tables]
        yield get_table_details(custom_table_names)

    @dlt.resource(write_disposition="merge", primary_key="ROLLNAME")
    def column_details():
        """Get metadata about table columns in SAP tables."""
        params = {
            "QUERY_TABLE": "DD04T",
            "FIELDS": ["ROLLNAME", "DDTEXT"],
            "OPTIONS": [{"TEXT": "DDLANGUAGE = 'EN'"}],
            "DELIMITER": delimiter,
        }
        logger.info("Extracting table columns...")
        yield get_response(params)

    return [
        standard_tables,
        standard_tables_details,
        custom_tables,
        custom_tables_details,
        column_details,
    ]


if __name__ == "__main__":
    pipeline = dlt.pipeline(
        pipeline_name="sap", destination="duckdb", dataset_name="bronze"
    )
    # pipeline.run(sap.with_resources("column_details", "custom_tables_details"))
    pipeline.run(sap().with_resources("column_details", "standard_tables_details"))



"""Playground for Qlik Sense APIs exploration.

We connect to two APIs:
- Qlik Sense Repository Service (QRS) API
- Qlik Engine JSON API

We connect using the default Virtual Proxy Service (built into Qlik Sense Enterprise
default configuration, so no server-side configuration is required).

For auth, for the time being we use the Windows login method (NTLM), combined with the
"header" method of authenticating to these APIs.

In the case of the QRS API, we use the requests library to make HTTP requests, and use
the requests_ntlm library to handle NTLM authentication.

For the Qlik Engine JSON API, we use the websocket library to create a WebSocket
connection. To authenticate, we use a Qlik session ID that we get from calling a
QRS API endpoint (the endpoint doesn't matter, we only care that the server returns
a session ID in its response) using the method described above.
"""

from collections.abc import Generator
import json
import secrets
import ssl
import string
from typing import Any

import dlt
from dlt.extract.resource import DltResource
from dotenv import load_dotenv
from loguru import logger
import requests
from requests_ntlm import HttpNtlmAuth
import websocket


load_dotenv()

# # Disable SSL warnings (optional but recommended for self-signed certs)
requests.packages.urllib3.disable_warnings()


def get_qrs_session(ntlm_username: str, ntlm_password: str) -> requests.Session:
    session = requests.Session()
    # Probably a typo of xsrf; anyway, this is a random string of 16 alphanumeric chars.
    alphabet = string.ascii_uppercase + string.ascii_lowercase + string.digits
    xrfkey = "".join(secrets.choice(alphabet) for _ in range(16))
    headers = {
        "x-qlik-xrfkey": xrfkey,
        "User-Agent": "Windows",
    }
    session.auth = HttpNtlmAuth(ntlm_username, ntlm_password)
    session.headers = headers
    session.params = {"xrfkey": xrfkey}
    session.verify = False
    return session


def get_qlik_session_id(
    qrs_api_base_url: str, ntlm_username: str, ntlm_password: str
) -> str:
    """Retrieve Qlik session ID from the QRS API.

    Raises:
        ValueError: If session ID could not be retrieved.

    Returns:
        str: The Qlik session ID.
    """
    session = get_qrs_session(ntlm_username=ntlm_username, ntlm_password=ntlm_password)

    logger.debug("Retrieving Qlik session ID...")

    full_url = qrs_api_base_url + "/about"
    response = session.get(full_url)
    response.raise_for_status()
    session_id = session.cookies.get("X-Qlik-Session")

    if not session_id:
        msg = "Could not retrieve Qlik session ID. Perhaps Qlik API's logic changed?"
        raise ValueError(msg)

    logger.debug(f"Retrieved Qlik session ID: {session_id}.")
    return session_id


def get_socket(engine_api_url: str, session_id: str) -> websocket.WebSocket:
    """Create a WebSocket connection to the Qlik Engine JSON API.

    Args:
        session (requests.Session): An already configured requests Session object.

    Returns:
        websocket.WebSocket: An authenticated socket connection.
    """
    # Get a Qlik session ID from QRS API.
    # This way, we can (indirectly) use header auth in the Qlik Engine JSON API.
    headers = {"Cookie": f"X-Qlik-Session={session_id}"}

    # Connect to the Qlik Engine JSON API.
    logger.info("Creating a socket...")

    socket = websocket.WebSocket(sslopt={"cert_reqs": ssl.CERT_NONE})
    socket.connect(engine_api_url, header=headers)
    on_authentication_msg = json.loads(socket.recv())
    logger.debug("On authentication message: ", on_authentication_msg)
    if on_authentication_msg["params"]["mustAuthenticate"]:
        msg = "Could not authenticate to Qlik Engine JSON API."
        msg += " Please ensure you're authenticated and provide a valid session ID."
        raise ValueError(msg)

    on_connected_msg = socket.recv()
    logger.debug("On connected message: ", on_connected_msg)

    return socket


def get(
    request: dict[str, Any],
    engine_api_url: str,
    session_id: str | None = None,
    socket: websocket.WebSocket | None = None,
) -> dict[str, Any]:
    """Retrieve a response from the Qlik Engine JSON API using WebSocket.

    Args:
        request (dict[str, Any]): The request body.
        engine_api_url (str): The URL of the Qlik Engine JSON API.
        session_id (str | None, optional): The ID of the Qlik session to use. Defaults
            to None.
        socket (websocket.WebSocket | None, optional): The socket to use. Defaults to
            None.

    Raises:
        ValueError: If neither a session nor a socket is provided, or if the socket is
            specified, but without its session.

    Returns:
        dict[str, Any] | None: The response from Qlik Engine JSON API.
    """
    if not session_id and not socket:
        msg = "Either `socket` or `session_id` must be provided."
        raise ValueError(msg)

    is_socket_externally_managed = bool(socket)
    if not is_socket_externally_managed:
        socket = get_socket(engine_api_url, session_id=session_id)

    logger.info(f"Calling '{request['method']}' method...")
    socket.send(json.dumps(request))
    response = json.loads(socket.recv())

    if not is_socket_externally_managed:
        socket.close()

    if machine_readable_error := response.get("error"):
        error_code = machine_readable_error["code"]
        error_message_short = machine_readable_error["message"]
        error_message_long = machine_readable_error["parameter"]
        human_readable_error = (
            f"Error {error_code} ('{error_message_short}'). " + error_message_long + "."
        )
        raise ValueError(human_readable_error)

    return response["result"]


@dlt.source
def qlik_sense(
    qrs_api_base_url: str = dlt.secrets.value,
    engine_api_url: str = dlt.secrets.value,
    ntlm_username: str = dlt.secrets.value,
    ntlm_password: str = dlt.secrets.value,
):
    # We use a single Qlik session for all requests.
    qlik_session_id = get_qlik_session_id(
        qrs_api_base_url, ntlm_username=ntlm_username, ntlm_password=ntlm_password
    )

    @dlt.resource(primary_key="qDocId", write_disposition="merge")
    def apps_engine(
        modified_at=dlt.sources.incremental(
            "modifiedDate", initial_value="2024-01-01T00:00:00Z"
        ),
    ) -> DltResource:
        """Get metadata about all apps in Qlik Sense from the Engine JSON API."""
        query = {
            "jsonrpc": "2.0",
            # "id": 1,
            "method": "GetDocList",
            "handle": -1,
            "params": [],
        }
        apps_info_nested = get(
            query, session_id=qlik_session_id, engine_api_url=engine_api_url
        )
        # Unnest and return only the modified apps.
        yield from [
            {
                "qDocId": app["qDocId"],
                "qDocName": app["qDocName"],
                "description": app["qMeta"]["description"],
                "createdDate": app["qMeta"]["createdDate"],
                "modifiedDate": app["qMeta"]["modifiedDate"],
                "stream": {
                    "id": app["qMeta"]["stream"]["id"]
                    if app["qMeta"].get("stream")
                    else None,
                    "name": app["qMeta"]["stream"]["name"]
                    if app["qMeta"].get("stream")
                    else None,
                },
            }
            for app in apps_info_nested["qDocList"]
            if app["qMeta"]["modifiedDate"] > modified_at.start_value
        ]

    @dlt.transformer(data_from=apps_engine, primary_key="id", write_disposition="merge")
    def app_details_qrs(apps: dict[str, Any]) -> Generator[dict[str, Any], None, None]:
        app_id = apps["qDocId"]
        logger.debug(f"Retrieving app {app_id} details...")

        session = get_qrs_session(
            ntlm_username=ntlm_username, ntlm_password=ntlm_password
        )
        full_url = qrs_api_base_url + f"/app/{app_id}"
        response = session.get(full_url)
        response.raise_for_status()
        yield response.json()

    @dlt.transformer(data_from=apps_engine, primary_key="id", write_disposition="merge")
    def app_details_engine(
        apps: dict[str, Any],
    ) -> Generator[dict[str, Any], None, None]:
        app_id = apps["qDocId"]
        # These two requests are connected, so we need to perform them on the same
        # socket.
        socket = get_socket(engine_api_url, session_id=qlik_session_id)
        open_app_query = {
            "jsonrpc": "2.0",
            "id": 1,
            "method": "OpenDoc",
            "handle": -1,
            "params": {"qDocName": app_id},
        }
        lineage_query = {
            "jsonrpc": "2.0",
            "id": 2,
            "method": "GetLineage",
            "handle": 1,
            "params": {},
        }
        try:
            _ = get(open_app_query, socket=socket, engine_api_url=engine_api_url)
            lineage = get(lineage_query, socket=socket, engine_api_url=engine_api_url)[
                "qLineage"
            ]
        except Exception as e:
            msg = f"Failed retrieving lineage for app: {app_id}."
            raise ValueError(msg) from e
        finally:
            socket.close()

        # Add app ID to the table since it's not in the response.
        for row in lineage:
            row["id"] = app_id
            # row["models"] = extract_models(row)  # TODO
        yield lineage

    return [apps_engine, app_details_qrs, app_details_engine]


if __name__ == "__main__":
    for app in qlik_sense().add_limit(1):
        print(json.dumps(app, indent=4))



import dlt

from luma.metadata.sources import qlik_sense


pipe = dlt.pipeline(
    pipeline_name="qlik_to_duckdb",
    destination=dlt.destinations.duckdb("db.duckdb"),
    dataset_name="qlik_sense",
)
# load_package = pipe.run(qlik().add_limit(5), refresh="drop_data") # For testing.
load_package = pipe.run(qlik_sense().add_limit(5))
print(load_package)



"""A command-line interface (CLI) for the Luma application.

Provides commands for database operations, configuration management, and more.
"""

from enum import Enum
from importlib import import_module
import importlib.metadata
import json
import sys
import time
from typing import Union

from loguru import logger
import requests
from rich import print
from rich.console import Console
import typer
import urllib3

from luma.commands import config, dbt, postgres
from luma.utils import (
    check_ingestion_results,
    check_ingestion_status,
    perform_ingestion_request,
)
from luma.utils.options import (
    DryRun,
    Follow,
    FollowTimeout,
    IngestionId,
    LumaURL,
)
from luma.utils.state import state


# Set the logging level to INFO by default.
logger.remove()
logger.add(sys.stdout, level="INFO")


__version__ = importlib.metadata.version(__package__ or __name__)

# Disable warnings related to insecure requests for specific cases
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

console = Console()


class MetadataSource(str, Enum):
    """Available metadata sources."""

    POWERBI = "powerbi"
    QLIK_SENSE = "qlik_sense"
    SAP = "sap"


class IngestionStatus(Enum):
    successful = 0
    failed = 1
    pending = 2


# Create a Typer application with configured properties
app = typer.Typer(
    name="luma",
    no_args_is_help=True,
    pretty_exceptions_enable=True,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_short=True,
)


def version_callback(show_version: bool) -> str | None:
    """Print the version of the application.

    Args:
        show_version (bool): If True, shows the version.
    """
    if show_version:
        typer.echo(__version__)
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(  # noqa: ARG001
        None,
        "--version",
        "-v",
        callback=version_callback,
        is_eager=True,
        help="Show the version and exit.",
    ),
    config_dir: str = typer.Option(
        None,
        "--config-dir",
        "-c",
        envvar="LUMA_CONFIG_DIR",
        help="The directory containing the Luma configuration file.",
    ),
) -> None:
    """Main function for the Typer application.

    Args:
        version (bool): Flag to show the version and exit.
        config_dir (str): The directory containing the Luma configuration file.
    """
    state["config_dir"] = config_dir


@app.command()
def status(
    ingestion_id: IngestionId,
    luma_url: str = LumaURL,
) -> Union[str, dict]:
    """Retrieve the status of an ingestion."""
    results = check_ingestion_results(luma_url, ingestion_id)
    print(f"Ingestion results for ID {ingestion_id}:")
    print(results)
    return results


@app.command()
def ingest(
    source: MetadataSource,
    luma_url: str = LumaURL,
    dry_run: bool = DryRun,
    follow: bool = Follow,
    follow_timeout: int = FollowTimeout,
) -> tuple[list[requests.Response], list[str | None]]:
    """Ingest metadata from external sources into Luma."""
    endpoint = f"{luma_url}/api/v1/ingest/dashboard"
    module = import_module(f"luma.metadata.sources.{source.value}.pipeline")

    # Retrieve generated manifest(s) and send them to Luma.
    manifests = module.pipeline()
    responses = []
    ingestion_ids = []
    for manifest in manifests:
        payload = json.loads(manifest.json(by_alias=True))
        n_items = len(payload["payload"])

        logger.debug(
            f"Sending {n_items} items from {source.value} metadata source to Luma instance {luma_url}..."
        )

        # If in dry run mode, print the bundle and exit.
        if dry_run:
            print(payload)
            raise typer.Exit(0)

        # Send ingestion request.
        response, ingestion_id = perform_ingestion_request(
            url=endpoint, payload=payload
        )
        if not response.ok:
            raise typer.Exit(1)

        if follow and ingestion_id:
            ingestion_status = None

            with console.status("Waiting...", spinner="dots"):
                for _ in range(follow_timeout):
                    ingestion_status = check_ingestion_status(luma_url, ingestion_id)
                    if ingestion_status == IngestionStatus.successful.value:
                        response = check_ingestion_results(luma_url, ingestion_id)
                        print()
                        print(f"Ingestion results for ID {ingestion_id}:")
                        print()
                        print(response)

                    if ingestion_status == IngestionStatus.failed.value:
                        print()
                        print(f"Ingestion failed for ID {ingestion_id}")

                    if ingestion_status == IngestionStatus.pending.value:
                        time.sleep(1)

            if ingestion_status != IngestionStatus.successful.value:
                status_human_readable = IngestionStatus(ingestion_status).name
                print(
                    f"Ingestion did not complete successfully within {follow_timeout} seconds. Status: {status_human_readable}."
                )

        responses.append(response)
        ingestion_ids.append(ingestion_id)

    return responses, ingestion_ids


# Add commands to the Typer application
app.add_typer(dbt.app, name="dbt", help="Ingest metadata from dbt.")
app.add_typer(postgres.app, name="postgres", help="Ingest metadata from Postgres.")
app.add_typer(config.app, name="config", help="Manage Luma instance configuration.")



"""Extract a database table manifest from SAP metadata.

We use the local DuckDB instance to join the two source tables in order to be able
to produce models expected by Luma.

Manifests are saved in batches (50,000 tables per batch) in order to limit payload size.
"""

from itertools import batched

import dlt
from loguru import logger

from luma.metadata.models.database import (
    DatabaseTable,
    DatabaseTableColumn,
    DatabaseTableManifest,
    DatabaseTableSchemaMetadata,
)


pipeline = dlt.pipeline(
    pipeline_name="sap", destination="duckdb", dataset_name="bronze"
)

custom_tables_query = """
SELECT tabname, fieldname, datatype, leng, ddtext
FROM custom_tables_details
LEFT JOIN column_details ON custom_tables_details.fieldname = column_details.rollname
ORDER BY tabname
"""

standard_tables_query = """
SELECT tabname, fieldname, datatype, leng, ddtext
FROM standard_tables_details
LEFT JOIN column_details ON standard_tables_details.fieldname = column_details.rollname
ORDER BY tabname
"""


def transform():
    with (
        pipeline.sql_client() as client,
        client.execute_query(standard_tables_query) as cursor,
    ):
        table_to_columns = {}
        # Create a table -> columns mapping.
        for i, df in enumerate(cursor.iter_df(chunk_size=50000)):
            logger.info(
                f"Extracting table->column mappings from chunk {i} ({len(df)} rows)..."
            )
            for _, row in df.iterrows():
                table_name = row["tabname"]
                column = DatabaseTableColumn(
                    name=row["fieldname"],
                    type=row["datatype"],
                    length=str(row["leng"]),
                    description=row["ddtext"],
                )
                if table_name not in table_to_columns:
                    table_to_columns[table_name] = {"columns": []}
                table_to_columns[table_name]["columns"].append(column)

    # Create table manifests from the mapping.
    tables = (
        DatabaseTable(
            name=table_name, columns=table_to_columns[table_name]["columns"], type="sap"
        )
        for table_name in table_to_columns
    )
    batch_size = 50000
    for i, table_batch in enumerate(batched(tables, batch_size)):
        logger.info(f"Generating manifest for table batch {i}...")

        manifest = DatabaseTableManifest(
            metadata=DatabaseTableSchemaMetadata(schema="database_table", version=1),
            payload=list(table_batch),
        )

        yield manifest

        # logger.info(f"Writing {len(manifest.payload)} tables to batch {i} manifest...")

        # with open(f"database_table_manifest__batch_{i}.json", "w") as f:
        #     f.write(manifest.json())



from datetime import datetime
from typing import Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field


class DBTMetadata(BaseModel):
    dbt_version: Optional[str]
    dbt_schema_version: Optional[str]
    generated_at: Optional[str]
    adapter_type: Optional[str]
    dbt_env: Optional[dict]
    invocation_id: Optional[str]
    send_anonymous_usage_stats: Optional[bool]
    project_id: Optional[str]
    user_id: Optional[str]


Meta = Dict[str, Union[str, int, float, bool, List[str], List[dict]]]


class Column(BaseModel):
    name: Optional[str]
    description: Optional[str]
    meta: Meta = {}
    data_type: Optional[str]
    quote: Optional[str]
    tags: Optional[List[str]]


class Macro(BaseModel):
    # Common
    name: Optional[str] = Field(description="Name of the object")
    unique_id: Optional[str]
    package_name: Optional[str]
    root_path: Optional[str]
    path: Optional[str]
    original_file_path: Optional[str]

    # Important
    resource_type: Optional[Literal["macro"]]
    meta: Meta = {}
    description: Optional[str]
    tags: Optional[list]

    # Other
    macro_sql: Optional[str]
    depends_on: Optional[dict]
    docs: Optional[dict]
    patch_path: Optional[str]
    arguments: Optional[list]
    created_at: Optional[float]


class Doc(BaseModel):
    __tablename__: str = "docs"

    # Common
    name: Optional[str] = Field(description="Name of the object")
    unique_id: Optional[str]
    package_name: Optional[str]
    root_path: Optional[str]
    path: Optional[str]
    original_file_path: Optional[str]

    # Other
    block_contents: Optional[str]


class Exposure(BaseModel):
    __tablename__: str = "exposure"

    # Common
    name: Optional[str] = Field(description="Name of the object")
    unique_id: Optional[str]
    package_name: Optional[str]
    root_path: Optional[str]
    path: Optional[str]
    original_file_path: Optional[str]


class Metrics(BaseModel):
    __tablename__: str = "metric"

    # Common
    name: Optional[str] = Field(description="Name of the object")
    unique_id: Optional[str]
    package_name: Optional[str]
    root_path: Optional[str]
    path: Optional[str]
    original_file_path: Optional[str]


class SourceBase(BaseModel):
    __tablename__: str = "sources"

    # Common
    name: Optional[str] = Field(description="Name of the object")
    unique_id: Optional[str]
    package_name: Optional[str]
    root_path: Optional[str]
    path: Optional[str]
    original_file_path: Optional[str]

    # Important
    resource_type: Optional[str]
    relation_name: Optional[str]
    description: Optional[str]
    loaded_at_field: Optional[str]

    # Other
    schema_field: Optional[str] = Field(alias="schema")
    source_name: Optional[str]
    source_description: Optional[str]
    loader: Optional[str]
    identifier: Optional[str]
    created_at: Optional[float]


class Source(SourceBase):
    meta: Meta = {}
    columns: Optional[Dict[str, Column]]
    resource_type: Optional[Literal["source"]]
    tags: Optional[list]
    freshness: Optional[dict]
    database: Optional[str]
    fqn: Optional[list]
    quoting: Optional[dict]
    external: Optional[dict]
    source_meta: Optional[dict]
    config: Optional[dict]
    patch_path: Optional[dict]
    unrendered_config: Optional[dict]


class Node(BaseModel):
    """
    Fields common for every Node (models, seeds, tests).
    """

    compiled: Optional[bool]
    database: Optional[str]
    schema_field: Optional[str] = Field(alias="schema")
    fqn: Optional[list]
    unique_id: str
    raw_code: Optional[str]
    language: Optional[str]
    package_name: str
    root_path: Optional[str]
    path: str
    original_file_path: str
    name: str
    resource_type: str
    alias: str
    checksum = dict
    config: Optional[dict]
    tags: Optional[list]
    refs: Optional[list]
    sources: Optional[dict]
    metrics: Optional[dict]
    depends_on: Optional[dict]
    description: Optional[str]
    columns: Optional[Dict[str, Column]]
    meta: Meta = {}
    docs: Optional[dict]
    patch_path: Optional[str]
    compiled_path: Optional[str]
    build_path: Optional[str]
    deferred: Optional[bool]
    unrendered_config: Optional[dict]
    created_at: Optional[float]
    config_call_dict: Optional[dict]
    compiled_code: Optional[str]
    extra_ctes_injected: Optional[bool]
    extra_ctes: Optional[dict]
    relation_name: Optional[str]
    catalog_created_at: Optional[datetime]
    catalog_updated_at: Optional[datetime]


class Model(Node):
    pass


class Seed(Node):
    pass


class Test(Node):
    pass


class ManifestDict(BaseModel):
    dbt_metadata: Optional[DBTMetadata] = Field(alias="metadata")
    nodes: Optional[Dict[str, Node]]
    sources: Optional[Dict[str, Source]]
    macros: Optional[Dict[str, Macro]]
    docs: Optional[Dict[str, Doc]]
    exposures: Optional[Dict[str, Exposure]]
    metrics: Optional[Dict[str, Metrics]]
    selectors: Optional[Dict[str, str]]
    disabled: Optional[dict]
    parent_map: Optional[Dict[str, List[Optional[str]]]]
    child_map: Optional[Dict[str, List[Optional[str]]]]


class CatalogNodeMetadata(BaseModel):
    _type: str = Field(alias="type")
    schema_field: Optional[str] = Field(alias="schema")
    name: str
    database: Optional[str]
    comment: Optional[str]
    owner: Optional[str]


class CatalogNodeColumns(BaseModel):
    type: Optional[str]
    _index: int = Field(alias="index")
    name: Optional[str]
    comment: Optional[str]


class CatalogNodeStats(BaseModel):
    id: Optional[str]
    label: Optional[str]
    value: Optional[Union[bool, str, float]]
    include: Optional[bool]
    description: Optional[str]


class CatalogNodes(BaseModel):
    dbt_metadata: CatalogNodeMetadata = Field(alias="metadata")
    columns: Dict[str, CatalogNodeColumns]
    stats: CatalogNodeStats
    unique_id: Optional[str]


class CatalogSource(CatalogNodes):
    pass


class CatalogDict(BaseModel):
    dbt_metadata: DBTMetadata = Field(alias="metadata")
    nodes: Dict[str, CatalogNodes]
    sources: Dict[str, CatalogSource]
    errors: Optional[str]


class RunResultsTimingInfo(BaseModel):
    name: str
    started_at: Optional[str]
    completed_at: Optional[str]


class RunResultsResults(BaseModel):
    status: str
    timing: List[RunResultsTimingInfo]
    thread_id: str
    execution_time: float
    adapter_response: dict
    message: Optional[str]
    failures: Optional[int]
    unique_id: str


class RunResultsDict(BaseModel):
    dbt_metadata: DBTMetadata = Field(alias="metadata")
    results: List[dict]
    elapsed_time: float
    args: Optional[dict]


class SourcesFreshnessRuntimeError(BaseModel):
    unique_id: str
    error: Optional[Union[str, int]]
    status: str


class SourceFreshnessThresholdWarnAfter(BaseModel):
    count: Optional[int]
    period: Optional[str]


class SourceFreshnessThresholdErrorAfter(SourceFreshnessThresholdWarnAfter):
    pass


class SourceFreshnessThreshold(BaseModel):
    warn_after: Optional[SourceFreshnessThresholdWarnAfter]
    error_after: Optional[SourceFreshnessThresholdErrorAfter]
    filter: Optional[str]


class SourcesTimingInfo(RunResultsTimingInfo):
    pass


class SourcesFreshnessOutput(BaseModel):
    unique_id: str
    max_loaded_at: str
    snapshotted_at: str
    max_loaded_at_time_ago_in_s: float
    status: str
    criteria: SourceFreshnessThreshold
    adapter_response: dict
    timing: SourcesTimingInfo
    thread_id: str
    execution_time: float


class SourcesDict(BaseModel):
    dbt_metadata: DBTMetadata = Field(alias="metadata")
    results: List[Union[SourcesFreshnessRuntimeError, SourcesFreshnessOutput]]
    elapsed_time: float


class BundleJson(BaseModel):
    manifest_json: Optional[ManifestDict]
    catalog_json: Optional[CatalogDict]
    sources_json: Optional[SourcesDict]
    run_results_json: Optional[RunResultsDict]



import json
from pathlib import Path
import re

from luma.metadata.models.bi import (
    Dashboard,
    DashboardManifest,
    DashboardSchemaMetadata,
    DataModel,
)
from luma.metadata.sources.powerbi.models import WorkspaceInfo


def transform(workspace_info: WorkspaceInfo) -> DashboardManifest:
    # Extract tables from the Power BI metadata.
    tables = extract_tables(workspace_info)
    reports = extract_reports(workspace_info, tables=tables)

    return DashboardManifest(
        metadata=DashboardSchemaMetadata(schema="dashboard", version=1),
        payload=reports,
    )


def extract_tables(workspace_info) -> list[dict]:
    tables = []
    # Each dataset table can only have one database table as a source.
    for workspace in workspace_info.workspaces:
        for dataset in workspace.datasets:
            for dataset_table in dataset.tables:
                if dataset_table.source is None:
                    continue

                # Extract the underlying database table.
                source_expression = dataset_table.source[0].expression
                table_database_table = _extract_table_from_expression(source_expression)

                if not table_database_table:
                    continue

                database_table_name = table_database_table["name"]
                tables.append({
                    "dataset_id": dataset.id,
                    "dataset_table_name": dataset_table.name,
                    "database_table_name": database_table_name,
                    "database_table_schema": table_database_table.get("schema"),
                    "database_table_database": table_database_table.get("database"),
                    "columns": [column.name for column in dataset_table.columns],
                })
    return tables


def extract_reports(
    workspace_info: WorkspaceInfo, tables: list[dict]
) -> list[Dashboard]:
    reports = []
    for workspace in workspace_info.workspaces:
        for report in workspace.reports:
            # We're not interested in PowerBI Apps. Not sure why they're included
            #  - either way, the original report the app is based on is already included
            # in the response.
            if report.name.startswith("[App]"):
                continue

            report_filtered = {}
            report_id = report.id
            report_filtered["external_id"] = report_id
            report_filtered["url"] = (
                "https://app.powerbi.com/groups/" + workspace.id
                or "" + "/reports/" + report_id
            )
            report_filtered["type"] = "powerbi"
            report_filtered["name"] = report.name
            report_filtered["workspace"] = workspace.name
            report_filtered["created_at"] = report.createdDateTime
            report_filtered["modified_at"] = report.modifiedDateTime
            report_filtered["owners"] = [
                {
                    "user_id": user.graphId,
                    "username": user.identifier,
                    "name": user.displayName,
                }
                for user in report.users
                if user.reportUserAccessRight == "Owner"
            ]

            report_tables = [
                {
                    "name": table["database_table_name"],
                    "schema": table["database_table_schema"],
                    "database": table["database_table_database"],
                    "columns": table["columns"],
                }
                for table in tables
                if table["dataset_id"] == report.datasetId
            ]
            report_filtered["parent_models"] = report_tables

            reports.append(report_filtered)

    return reports


def _extract_table_from_expression(expression: str) -> DataModel | None:
    """Extract schema and table name from expression."""
    # Get database name.
    database_name_expr = re.search(
        r'AmazonRedshift\.Database\s*\(\s*".*?"\s*,\s*"([^" ]+)"\s*\)',
        expression,
        re.IGNORECASE,
    )
    if not database_name_expr:
        return None
    database_name = database_name_expr.group(1).strip()

    # Get schema names.
    schema_pattern = r'(\w+)\s*=\s*(?:Source|AmazonRedshift\.Database\(.*\))\s*{\[Name="([^" ]+)"\]}\[Data\]'
    schema_match = re.findall(schema_pattern, expression, re.IGNORECASE)

    if not schema_match:
        return None

    schema_name = schema_match[0][0].strip()
    schema_var_name = schema_match[0][1].strip()

    # Get table metadata.
    escaped_schema_var = re.escape(schema_var_name)
    table_pattern = (
        r"\w+\s*=\s*" + escaped_schema_var + r'{\[Name="([^" ]+)"\]}\[Data\]'
    )
    table_match = re.findall(table_pattern, expression, re.IGNORECASE)

    if not table_match:
        return None

    table = {
        "name": table_match[0].strip(),
        "schema": schema_name.strip(),
        "database": database_name,
    }

    return table


if __name__ == "__main__":
    with Path("powerbi_workspace_info.json", encoding="utf-8").open() as f:
        workspace_info = json.load(f)

    dashboard_manifest = transform(workspace_info)

    with Path("powerbi_extracted.json", encoding="utf-8").open("w") as f:
        f.write(dashboard_manifest.json(by_alias=True))
    # # print(json.dumps(tables, indent=4))



from luma.luma import app
from typer.testing import CliRunner

runner = CliRunner()


def test_ingest_v1_7(METADATA_DIR_V1_7):
    result = runner.invoke(
        app,
        [
            "dbt",
            "ingest",
            "--metadata-dir",
            METADATA_DIR_V1_7 / "model_metadata",
            "--luma-url",
            "http://localhost:8000",
            "--dry-run",
        ],
    )
    assert result.exit_code == 0



import click

from rich.console import Console
from typer.testing import CliRunner

from luma.luma import app
from luma.utils.luma_utils import get_config

runner = CliRunner()


def test_get_config(config_dir):
    """Test if the configuration is correctly retrieved from a given directory."""
    config = get_config(config_dir=config_dir)
    assert config
    assert config.groups is not None
    assert config.owners is not None


def test_show(config_dir):
    """Test if the 'show' command correctly displays the configuration."""
    result = runner.invoke(
        app,
        [
            "config",
            "show",
            "--config-dir",
            config_dir,
        ],
    )

    assert result.exit_code == 0
    config = get_config(config_dir=config_dir)
    console = Console(record=True)
    console.print(config)
    text = console.export_text()
    assert text in result.output


def test_send(config_dir, test_server):
    """Test if the 'send' command successfully sends the configuration."""
    result = runner.invoke(
        app,
        ["config", "send", "--config-dir", config_dir, "--luma-url", test_server],
    )

    assert result.exit_code == 0
    assert "The request was successful!" in result.output


# Invalid schema tests
def test_get_config_invalid_schema(config_dir_invalid_schema):
    """Test if the configuration retrieval correctly handles an invalid schema."""
    try:
        config = get_config(config_dir=config_dir_invalid_schema)
        assert config is None
    except click.exceptions.Abort:
        pass


def test_show_invalid_schema(config_dir_invalid_schema):
    """Test if the 'show' command correctly handles an invalid configuration schema."""
    result = runner.invoke(
        app,
        [
            "config",
            "show",
            "--config-dir",
            config_dir_invalid_schema,
        ],
    )

    assert result.exit_code == 1
    assert "Error parsing YAML file" in result.output


def test_send_invalid_schema(config_dir_invalid_schema, test_server):
    """Test if the 'send' command correctly handles an invalid configuration schema."""
    result = runner.invoke(
        app,
        [
            "config",
            "send",
            "--config-dir",
            config_dir_invalid_schema,
            "--luma-url",
            test_server,
        ],
    )

    assert result.exit_code == 1
    assert "Error parsing YAML file" in result.output
